// -*- C++ -*- (c) 2016 Jan Mrázek <email@honzamrazek.cz>
//                 2016 Vladimir Still <xstill@fi.muni.cz>

#ifndef __DIOS_FAULT_H__
#define __DIOS_FAULT_H__

#ifdef __cplusplus
#define EXTERN_C extern "C" {
#define CPP_END }
#endif

#ifdef __cplusplus

#include <algorithm>
#include <cstdarg>
#include <divine/metadata.h>
#include <cxxabi.h>

#include <dios/core/main.hpp>
#include <dios/core/trace.hpp>
#include <dios/core/syscall.hpp>

namespace __dios {

enum FaultFlag
{
    Enabled       = 0x01,
    Continue      = 0x02,
    UserSpec      = 0x04,
    AllowOverride = 0x08,
    Artificial    = 0x10
};

template < typename Next >
struct Fault: public Next {
    static constexpr int fault_count = _DiOS_SF_Last;

    struct Process : Next::Process
    {
        Process() {
            std::fill_n( faultConfig, _DiOS_SF_Last,
                FaultFlag::Enabled | FaultFlag::AllowOverride );
        }

        uint8_t faultConfig[ _DiOS_SF_Last ];
    };

    struct MallocNofailGuard {
    private:
        MallocNofailGuard( Fault< Next >& f )
            : _f( f ),
              _config( _f.getCurrentConfig() ),
              _s( _config[ _DiOS_SF_Malloc ] )
        {
            _config[ _DiOS_SF_Malloc ] = 0;
            _f.updateSimFail();
        }
    public:
        ~MallocNofailGuard() {
            _config[ _DiOS_SF_Malloc ] = _s;
            _f.updateSimFail();
        }

        Fault& _f;
        uint8_t *_config;
        uint8_t _s;

        friend Fault;
    };

    MallocNofailGuard makeMallocNofail() {
        return { *this };
    }

    Fault()
        : ready( false )
    {}

    template< typename Setup >
    void setup( Setup s ) {
        traceAlias< Fault >( "{Fault}" );
        __vm_control( _VM_CA_Set, _VM_CR_FaultHandler, handler< typename Setup::Context > );
        load_user_pref( s.proc1->faultConfig, s.opts );

        if ( extractOpt( "debug", "faultcfg", s.opts ) ) {
            trace_config( 1 );
            __vm_control( _VM_CA_Bit, _VM_CR_Flags, _VM_CF_Error, _VM_CF_Error );
        }
        ready = true;
        Next::setup( s );
    }

    void getHelp( Map< String, HelpOption >& options ) {
        const char *opt1 = "[force-]{ignore|report|abort}";
        if ( options.find( opt1 ) != options.end() ) {
            __dios_trace_f( "Option %s already present", opt1 );
            __dios_fault( _DiOS_F_Config, "Option conflict" );
        }
        options[ opt1 ] = { "configure fault, force disables program override",
          { "assert",
            "arithmetic",
            "memory",
            "control",
            "locking",
            "hypercall",
            "notimplemented",
            "diosassert" } };

        const char *opt2 = "{nofail|simfail}";
        if ( options.find( opt2 ) != options.end() ) {
            __dios_trace_f( "Option %s already present", opt2 );
            __dios_fault( _DiOS_F_Config, "Option conflict" );
        }
        options[ opt2 ] = { "enable/disable simulation of failure",
          { "malloc" } };

        Next::getHelp( options );
    }

    uint8_t *getCurrentConfig() {
        auto *thread = Next::getCurrentThread();
        if ( !thread )
            return nullptr;
        return static_cast< Process * >( thread->_proc )->faultConfig;
    }

    void updateSimFail( uint8_t *config = nullptr ) {
        config = config ? config : getCurrentConfig();
        uint8_t flags = 0;
        uint8_t bit = 1;
        for ( int x = _DiOS_SF_First; x != _DiOS_SF_Last; x++, bit <<= 1 )
            flags |= config[ x ] & FaultFlag::Enabled ? bit : 0;
        auto *meta = __md_get_global_meta( "_DiOS_SimFail_flags" );
        if ( !meta )
            traceInternal( 0, "Warning: Cannot update SimFail" );
        else
            *reinterpret_cast< char *>( meta->address ) = flags;
    }

    void backtrace( _VM_Frame * frame ) {
        auto guard = makeMallocNofail();
        char *buffer = static_cast< char * >( malloc( 1024 ) );
        size_t len = 1024;
        int i = 0;
        for ( auto *f = frame; f != nullptr; f = f->parent ) {
            const char *name = __md_get_pc_meta( f->pc )->name;
            int status;
            char *b = __cxxabiv1::__cxa_demangle( name, buffer, &len, &status );
            if ( b )
                traceInternal( 0, "  %d: %s", ++i, b );
            else
                traceInternal( 0, "  %d: %s", ++i, name );
        }
        free( buffer );
    }

    void fault_handler( int kernel, _VM_Frame * frame, int what )
    {
        if ( what >= fault_count ) {
            traceInternal( 0, "Unknown fault in handler" );
            Next::die();
        }

        auto *cfg = getCurrentConfig();
        // If no thread exists, trigger the fault
        uint8_t fault_cfg = cfg ? cfg[ what ] : FaultFlag::Enabled;
        if ( !ready || fault_cfg & FaultFlag::Enabled ) {
            if ( kernel )
                traceInternal( 0, "Fault in kernel: %s", fault_to_str( what ).c_str() );
            else
                traceInternal( 0, "Fault in userspace: %s", fault_to_str( what ).c_str() );
            __vm_control( _VM_CA_Bit, _VM_CR_Flags, _VM_CF_Error, _VM_CF_Error );
            traceInternal( 0, "Backtrace:" );
            backtrace( frame );

            if ( !ready || !( fault_cfg & FaultFlag::Continue ) )
                Next::die();
        }
    }

    template < typename Context >
    static void handler( _VM_Fault _what, _VM_Frame *cont_frame,
                         void (*cont_pc)(), ... ) noexcept
    {
        auto mask = reinterpret_cast< uintptr_t >(
            __vm_control( _VM_CA_Get, _VM_CR_Flags,
                          _VM_CA_Bit, _VM_CR_Flags, _VM_CF_Mask, _VM_CF_Mask ) ) & _VM_CF_Mask;
        int kernel = reinterpret_cast< uintptr_t >(
            __vm_control( _VM_CA_Get, _VM_CR_Flags ) ) & _VM_CF_KernelMode;
        auto *frame = static_cast< _VM_Frame * >( __vm_control( _VM_CA_Get, _VM_CR_Frame ) )->parent;

        if ( kernel ) {
            void *ctx = __vm_control( _VM_CA_Get, _VM_CR_State );
            auto& fault = *static_cast< Context * >( ctx );
            fault.fault_handler( kernel, frame, _what );
        }
        else {
            __dios_syscall( SYS_fault_handler, nullptr, kernel, frame, _what  );
        }

        // Continue if we get the control back
        __vm_control( _VM_CA_Set, _VM_CR_Frame, cont_frame,
                      _VM_CA_Set, _VM_CR_PC, cont_pc,
                      _VM_CA_Bit, _VM_CR_Flags, _VM_CF_Mask, mask ? _VM_CF_Mask : 0 );
        __builtin_unreachable();
    }

    static int str_to_fault( String fault ) {
        std::transform( fault.begin(), fault.end(), fault.begin(), ::tolower );
        if ( fault == "assert" )
            return _VM_F_Assert;
        if ( fault == "arithmetic" )
            return _VM_F_Arithmetic;
        if ( fault == "memory" )
            return _VM_F_Memory;
        if ( fault == "control" )
            return _VM_F_Control;
        if ( fault == "locking" )
            return _VM_F_Locking;
        if ( fault == "hypercall" )
            return _VM_F_Hypercall;
        if ( fault == "notimplemented" )
            return _VM_F_NotImplemented;
        if ( fault == "diosassert" )
            return static_cast< _VM_Fault >( _DiOS_F_Assert );
        if ( fault == "diosconfig" )
            return static_cast< _VM_Fault >( _DiOS_F_Config );
        if ( fault == "malloc" )
            return static_cast< _VM_Fault >( _DiOS_SF_Malloc );
        return static_cast< _VM_Fault >( -1 );
    }

    static String fault_to_str( int f )  {
        switch( f ) {
            case _VM_F_Assert: return "assert";
            case _VM_F_Arithmetic: return "arithmetic";
            case _VM_F_Memory: return "memory";
            case _VM_F_Control: return "control";
            case _VM_F_Locking: return "locking";
            case _VM_F_Hypercall: return "hypercall";
            case _VM_F_NotImplemented: return "notimplemented";
            case _DiOS_F_Assert: return "diosassert";
            case _DiOS_F_Config: return "diosconfig";
            case _DiOS_SF_Malloc: return "malloc";
        }
        return "unknown";
    }

    void trace_config( int indent ) {
        auto config = getCurrentConfig();
        __dios_trace_i( indent, "fault and simfail configuration:" );
        for (int f = _VM_F_NoFault + 1; f != _DiOS_SF_Last; f++ ) {
            uint8_t cfg = config[f];
            String name = fault_to_str( f );
            String force = cfg & FaultFlag::AllowOverride ? "" : "force-";
            String state;
            if ( f >= _DiOS_F_Last )
                state = cfg & FaultFlag::Enabled ? "simfail" : "nofail";
            else {
                if ( cfg & FaultFlag::Enabled && cfg & FaultFlag::Continue )
                    state = "report";
                else if ( cfg & FaultFlag::Enabled )
                    state = "abort";
                else
                    state = "ignore";
            }
            String def = cfg & FaultFlag::UserSpec ? "user" : "default";

            __dios_trace_i( indent + 1, "%s: %s%s, %s", name.c_str(), force.c_str(),
                            state.c_str(), def.c_str() );;
        }
    }

    void load_user_pref( uint8_t *config, SysOpts& opts )  {
        SysOpts unused;
        for( const auto& i : opts ) {
            String cmd = i.first;
            String f = i.second;

            uint8_t cfg = FaultFlag::AllowOverride | FaultFlag::UserSpec;
            const String force("force-");
            if ( cmd.size() > force.size() && std::equal( force.begin(), force.end(), cmd.begin() ) ) {
                cfg &= ~FaultFlag::AllowOverride;
                cmd = cmd.substr( force.size() );
            }

            bool simfail = false;
            if      ( cmd == "ignore" )  { }
            else if ( cmd == "report" )  { cfg |= FaultFlag::Enabled | FaultFlag::Continue; }
            else if ( cmd == "abort" )   { cfg |= FaultFlag::Enabled; }
            else if ( cmd == "nofail" )  { simfail = true; }
            else if ( cmd == "simfail" ) { cfg |= FaultFlag::Enabled; simfail = true; }
            else {
                unused.push_back( i );
                continue;
            };

            int f_num = str_to_fault( f );
            if ( f_num == - 1 ) {
                __dios_trace( 0, "Invalid argument '%s' in option '%s:%s'",
                    f.c_str(), cmd.c_str(), f.c_str() );
                __dios_fault( _DiOS_F_Config, "Invalid configuration of fault handler" );
            }

            if ( simfail && f_num < _DiOS_F_Last ) {
                __dios_trace( 0, "Invalid argument '%s' for command '%s'",
                    f.c_str(), cmd.c_str() );
                __dios_fault( _DiOS_F_Config, "Invalid configuration of fault handler" );
            }

            config[ f_num ] = cfg;
        }
        opts = unused;
        updateSimFail( config );
    }

    int _faultToStatus( int fault )  {
        auto config = getCurrentConfig();
        uint8_t cfg = config[ fault ];
        if ( fault < _DiOS_F_Last) { // Fault
            switch ( cfg & (FaultFlag::Enabled | FaultFlag::Continue) ) {
                case 0:
                    return _DiOS_FC_Ignore;
                case FaultFlag::Enabled:
                    return _DiOS_FC_Abort;
                case FaultFlag::Enabled | FaultFlag::Continue:
                    return _DiOS_FC_Report;
                default:
                    __builtin_unreachable();
            }
        }
        else { // Simfail
            if ( cfg | FaultFlag::Enabled )
                return _DiOS_FC_SimFail;
            else
                return _DiOS_FC_NoFail;
        }
    }

    int configure_fault( int fault, int cfg ) {
        auto config = getCurrentConfig();
        if ( fault >= fault_count ) {
            return _DiOS_FC_EInvalidFault;
        }

        uint8_t& fc = config[ fault ];
        if ( !( fc & FaultFlag::AllowOverride ) ) {
            return _DiOS_FC_ELocked;
        }

        int lastCfg = _faultToStatus( fault );
        if ( fault < _DiOS_F_Last) { // Fault
            switch( cfg ) {
                case _DiOS_FC_Ignore:
                    fc = FaultFlag::AllowOverride;
                    break;
                case _DiOS_FC_Report:
                    fc = FaultFlag::AllowOverride | FaultFlag::Enabled | FaultFlag::Continue;
                    break;
                case _DiOS_FC_Abort:
                    fc = FaultFlag::AllowOverride | FaultFlag::Enabled;
                    break;
                default:
                    return _DiOS_FC_EInvalidCfg;
            }
        }
        else { // Simfail
            switch( cfg ) {
                case _DiOS_FC_SimFail:
                    fc = FaultFlag::AllowOverride | FaultFlag::Enabled;
                    break;
                case _DiOS_FC_NoFail:
                    fc = FaultFlag::AllowOverride;
                    break;
                default:
                    return _DiOS_FC_EInvalidCfg;
            }
        }
        updateSimFail();
        return lastCfg;
    }

    int get_fault_config( int fault ) {
        if ( fault >= fault_count )
            return _DiOS_FC_EInvalidFault;
        return _faultToStatus( fault );
    }

    bool ready;
};

} // namespace __dios

#endif // __cplusplus

#endif // __DIOS_FAULT_H__
